import { interval, Observable, Observer, Subject } from 'rxjs';
import { distinctUntilChanged, share, takeWhile } from 'rxjs/operators';
import { WebSocketSubject, WebSocketSubjectConfig } from 'rxjs/webSocket';

const urljoin = require('url-join');

export interface RxWebSocketSubjectConfig<T> extends WebSocketSubjectConfig<T> {
  queryString?: string;
  reconnectInterval?: 5000;
  reconnectRetry?: 10;
}

export class RxWebSocketSubject<T> extends Subject<T> {
  private reconnectionObservable: Observable<number> | undefined;
  private wsSubjectConfig: WebSocketSubjectConfig<T>;
  private socket: WebSocketSubject<T> | undefined;
  private connectionObserver: Observer<boolean> | undefined;
  public connectionStatus: Observable<boolean>;

  private readonly reconnectInterval: number;
  private readonly reconnectRetry: number;

  constructor(private _rxConfig: RxWebSocketSubjectConfig<T>) {
    super();

    this.reconnectInterval =
      undefined !== this._rxConfig.reconnectInterval
        ? this._rxConfig.reconnectInterval
        : 5000;
    this.reconnectRetry =
      undefined !== this._rxConfig.reconnectRetry
        ? this._rxConfig.reconnectRetry
        : 10;

    this.connectionStatus = new Observable((observer: Observer<boolean>) => {
      this.connectionObserver = observer;
    }).pipe(
      share(),
      distinctUntilChanged(),
    );

    this.wsSubjectConfig = Object.assign({}, this._rxConfig);
    this.wsSubjectConfig.closeObserver = {
      next: (_e: CloseEvent) => {
        this.socket = undefined;
        if (undefined !== this.connectionObserver) {
          this.connectionObserver.next(false);
        }
      },
    };
    this.wsSubjectConfig.openObserver = {
      next: (_e: Event) => {
        if (undefined !== this.connectionObserver) {
          this.connectionObserver.next(true);
        }
      },
    };

    this.connect();
    this.connectionStatus.subscribe(isConnected => {
      if (
        !this.reconnectionObservable &&
        !isConnected
      ) {
        this.reconnect();
      }
    });
  }

  private connect(): void {
    this.wsSubjectConfig.url =
      undefined !== this._rxConfig.queryString
        ? urljoin(this._rxConfig.url, this._rxConfig.queryString)
        : this._rxConfig.url;
    this.socket = new WebSocketSubject(this.wsSubjectConfig);
    this.socket.subscribe(
      m => {
        this.next(m);
      },
      (_error: Event) => {
        if (!this.socket) {
          this.reconnect();
        }
      },
    );
  }

  private reconnect(): void {
    this.reconnectionObservable = interval(this.reconnectInterval).pipe(
      takeWhile((_v, index) => {
        return index < this.reconnectRetry && !this.socket;
      }),
    );
    this.reconnectionObservable.subscribe(
      () => {
        this.connect();
      },
      undefined,
      () => {
        this.reconnectionObservable = undefined;
        if (!this.socket) {
          this.complete();
          if (undefined !== this.connectionObserver) {
            this.connectionObserver.complete();
          }
        }
      },
    );
  }

  public disconnect(): void {
    if (undefined !== this.socket) {
      this.socket.complete();
    }
  }

  public send(data: T): void {
    if (undefined !== this.socket) {
      this.socket.next(data);
    }
  }
}
